//  sample_cvt.hpp: PCM sample converter

//  Copyright Takeshi Mouri 2006.
//  Use, modification, and distribution are subject to the
//  Boost Software License, Version 1.0. (See accompanying file
//  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef HAMIGAKI_AUDIO_SAMPLE_CVT_HPP
#define HAMIGAKI_AUDIO_SAMPLE_CVT_HPP

#include <boost/iostreams/categories.hpp>
#include <boost/iostreams/read.hpp>
#include <boost/iostreams/write.hpp>
#include <boost/cstdint.hpp>

namespace hamigaki { namespace audio {

namespace detail
{

struct cvt_8
{
    static boost::int32_t decode(const char* s)
    {
        boost::int32_t tmp = static_cast<unsigned char>(*s);
        return (tmp - 128) * static_cast<boost::int32_t>(16777216L);
    }

    static void encode(char* s, boost::int32_t n)
    {
        boost::uint8_t tmp = static_cast<boost::uint8_t>(
            n/static_cast<boost::int32_t>(16777216L) + 128);
        *s = static_cast<char>(static_cast<unsigned char>(tmp));
    }
};

struct cvt_16
{
    static boost::int32_t decode(const char* s)
    {
        boost::uint16_t tmp = static_cast<unsigned char>(s[0]);
        tmp |= static_cast<boost::uint16_t>(
            static_cast<unsigned char>(s[1])) << 8;

        boost::int32_t val = (tmp & 0x8000)
            ? -static_cast<boost::int32_t>(~tmp) - 1
            : static_cast<boost::int32_t>(tmp);

        return val * static_cast<boost::int32_t>(65536L);
    }

    static void encode(char* s, boost::int32_t n)
    {
        boost::int16_t val = static_cast<boost::int16_t>(
            n/static_cast<boost::int32_t>(65536L));

        boost::uint32_t tmp = (val > 0)
            ? static_cast<boost::uint32_t>(val)
            : 0x10000 - static_cast<boost::uint32_t>(-val);

        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>(tmp & 0xFF));
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>((tmp >> 8) & 0xFF));
    }
};

struct cvt_24
{
    static boost::int32_t decode(const char* s)
    {
        boost::uint32_t tmp = static_cast<unsigned char>(s[0]);
        tmp |= static_cast<boost::uint32_t>(
            static_cast<unsigned char>(s[1])) << 8;
        tmp |= static_cast<boost::uint32_t>(
            static_cast<unsigned char>(s[2])) << 16;

        boost::int32_t val = (tmp & 0x800000)
            ? -static_cast<boost::int32_t>(~tmp) - 1
            : static_cast<boost::int32_t>(tmp);

        return val * 256;
    }

    static void encode(char* s, boost::int32_t n)
    {
        boost::int32_t val = static_cast<boost::int32_t>(n/256);

        boost::uint32_t tmp = (val > 0)
            ? static_cast<boost::uint32_t>(val)
            : 0x1000000 - static_cast<boost::uint32_t>(-val);

        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>(tmp & 0xFF));
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>((tmp >> 8) & 0xFF));
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>((tmp >> 16) & 0xFF));
    }
};

struct cvt_32
{
    static boost::int32_t decode(const char* s)
    {
        boost::uint32_t tmp = static_cast<unsigned char>(s[0]);
        tmp |= static_cast<boost::uint32_t>(
            static_cast<unsigned char>(s[1])) << 8;
        tmp |= static_cast<boost::uint32_t>(
            static_cast<unsigned char>(s[2])) << 16;
        tmp |= static_cast<boost::uint32_t>(
            static_cast<unsigned char>(s[3])) << 24;

        boost::int32_t val = (tmp & 0x80000000)
            ? -static_cast<boost::int32_t>(~tmp) - 1
            : static_cast<boost::int32_t>(tmp);

        return val;
    }

    static void encode(char* s, boost::int32_t n)
    {
        boost::int32_t val = static_cast<boost::int32_t>(n/256);

        boost::uint32_t tmp = (val > 0)
            ? static_cast<boost::uint32_t>(val)
            : 0x1000000 - static_cast<boost::uint32_t>(-val);

        *(s++) = '\0';
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>(tmp & 0xFF));
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>((tmp >> 8) & 0xFF));
        *(s++) = static_cast<char>(
            static_cast<boost::uint8_t>((tmp >> 16) & 0xFF));
    }
};

} // namespace detail

class sample_cvt
{
public:
    typedef char char_type;

#if 1
    typedef boost::iostreams::multichar_output_filter_tag category;
#else
    struct category
        : boost::iostreams::dual_use
        , boost::iostreams::filter_tag
        , boost::iostreams::multichar_tag
    {};
#endif

    sample_cvt(std::streamsize from, std::streamsize to)
        : from_size_(from), to_size_(to)
    {
    }

    template<typename Source>
    std::streamsize read(Source& src, char* s, std::streamsize n)
    {
        if (from_size_ == to_size_)
            return boost::iostreams::read(src, s, n);

        std::streamsize total = 0;
        while (n != 0)
        {
            char buf[4];
            boost::iostreams::read(src, buf, from_size_);

            boost::int32_t val = decode(buf);
            encode(s, val);

            s += from_size_;
            n -= from_size_;
            total += from_size_;
        }
        return (total != 0) ? total : -1;
    }

    template<typename Sink>
    std::streamsize write(Sink& sink, const char* s, std::streamsize n)
    {
        if (from_size_ == to_size_)
            return boost::iostreams::write(sink, s, n);

        std::streamsize total = 0;
        while (n != 0)
        {
            boost::int32_t val = decode(s);
            s += from_size_;
            n -= from_size_;
            total += from_size_;

            char buf[4];
            encode(buf, val);
            boost::iostreams::write(sink, buf, to_size_);
        }
        return (total != 0) ? total : -1;
    }

private:
    std::streamsize from_size_;
    std::streamsize to_size_;

    boost::int32_t decode(const char* s)
    {
        if (from_size_ == 1)
            return detail::cvt_8::decode(s);
        else if (from_size_ == 2)
            return detail::cvt_16::decode(s);
        else if (from_size_ == 3)
            return detail::cvt_24::decode(s);
        else
            return detail::cvt_32::decode(s);
    }

    void encode(char* s, boost::int32_t val)
    {
        if (to_size_ == 1)
            detail::cvt_8::encode(s, val);
        else if (to_size_ == 2)
            detail::cvt_16::encode(s, val);
        else if (to_size_ == 3)
            detail::cvt_24::encode(s, val);
        else
            detail::cvt_32::encode(s, val);
    }
};

} } // End namespaces audio, hamigaki.

#endif // HAMIGAKI_AUDIO_SAMPLE_CVT_HPP
