//  demultiplexer.hpp: stream demultiplexer

//  Copyright Takeshi Mouri 2006.
//  Use, modification, and distribution are subject to the
//  Boost Software License, Version 1.0. (See accompanying file
//  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef HAMIGAKI_IOSTREAMS_DEMULTIPLEXER_HPP
#define HAMIGAKI_IOSTREAMS_DEMULTIPLEXER_HPP

#include <boost/iostreams/categories.hpp>
#include <boost/iostreams/close.hpp>
#include <boost/iostreams/write.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/assert.hpp>
#include <boost/shared_ptr.hpp>
#include <algorithm>
#include <vector>

namespace hamigaki { namespace iostreams {

namespace detail
{

class demuxer_component_base
{
public:
    virtual ~demuxer_component_base() {}
    virtual std::streamsize write(const char* s, std::streamsize n) = 0;
    virtual void close() = 0;
};

template<class Sink>
class demuxer_component : public demuxer_component_base
{
public:
    demuxer_component(const Sink& sink, std::streamsize elm_size)
        : sink_(sink), element_size_(elm_size)
    {
    }

    std::streamsize write(const char* s, std::streamsize n) // virtual
    {
        if (n < element_size_)
            return -1;
        return boost::iostreams::write(sink_, s, element_size_);
    }

    void close() // virtual
    {
        boost::iostreams::close(sink_, BOOST_IOS::out);
    }

    const Sink* component() const
    {
        return &sink_;
    }

private:
    Sink sink_;
    std::streamsize element_size_;
};

class demultiplexer_impl
{
private:
    struct closer
    {
        void operator()(detail::demuxer_component_base& sink) const
        {
            sink.close();
        }
    };

public:
    demultiplexer_impl() : block_size_(0)
    {
    }

    template<typename T>
    const T* component(int n) const
    {
        typedef detail::demuxer_component<T> wrapper;

        if (const wrapper* ptr = dynamic_cast<const wrapper*>(&sinks_[n]))
            return ptr->component();
        else
            return 0;
    }

    template<typename Sink>
    void push(const Sink& sink, std::streamsize elm_size)
    {
        sinks_.push_back(new detail::demuxer_component<Sink>(sink, elm_size));
        block_size_ += elm_size;
    }

    std::streamsize write(const char* s, std::streamsize n)
    {
        if (sinks_.empty())
            return (n != 0) ? n : -1;

        std::streamsize total = 0;
        while (n != 0)
        {
            std::streamsize amt = write_once(s, n);
            if (amt < 0)
                break;
            s += amt;
            n -= amt;
            total += amt;
        }

        return (total != 0) ? total : -1;
    }

    void close()
    {
        std::for_each(sinks_.begin(), sinks_.end(), closer());
    }

    std::streamsize block_size() const
    {
        return block_size_;
    }

private:
    boost::ptr_vector<detail::demuxer_component_base> sinks_;
    std::streamsize block_size_;

    std::streamsize write_once(const char* s, std::streamsize n)
    {
        std::streamsize total = 0;
        for (std::size_t i = 0; i < sinks_.size(); ++i)
        {
            std::streamsize amt = sinks_[i].write(s, n);
            if (amt < 0)
                return -1;
            s += amt;
            n -= amt;
            total += amt;
        }
        return total;
    }
};

} // namespace detail

class demultiplexer
{
public:
    typedef char char_type;

    struct category :
        boost::iostreams::sink_tag,
        boost::iostreams::closable_tag {};

    demultiplexer() : pimpl_(new detail::demultiplexer_impl)
    {
    }

    template<typename T>
    const T* component(int n) const
    {
        return pimpl_->component<T>(n);
    }

    template<typename Sink>
    void push(const Sink& sink, std::streamsize elm_size)
    {
        pimpl_->push(sink, elm_size);
    }

    template<typename Sink>
    void push(const Sink& sink)
    {
        pimpl_->push(sink, sink.block_size());
    }

    std::streamsize write(const char* s, std::streamsize n)
    {
        return pimpl_->write(s, n);
    }

    void close()
    {
        pimpl_->close();
    }

    std::streamsize block_size() const
    {
        return pimpl_->block_size();
    }

private:
    boost::shared_ptr<detail::demultiplexer_impl> pimpl_;
};

} } // End namespaces iostreams, hamigaki.

#endif // HAMIGAKI_IOSTREAMS_DEMULTIPLEXER_HPP
